#pragma once

#include <string>

#include "httplib.h"
#include "hack/logger/logger.hpp"

#include "trs/utils/var.hpp"
#include "trs/libs/transaction.hpp"
#include "trs/libs/function_manager.hpp"
#include "trs/libs/inspector.hpp"
#include "trs/libs/database.hpp"
#include "trs/helpers/verification.hpp"
#include "trs/helpers/execute.hpp"

namespace trs
{
  class http : public httplib::Server
  {
    public:
      http() = default;
      ~http() = default;

    public:
      void init(std::string app_connection, std::string log_connection)
      {
        set_read_timeout(5, 0);
        set_write_timeout(5, 0);
        set_idle_interval(0, 1'000'000);
        set_payload_max_length(1024 * 1024 * 512); // 512MB
        set_CORS();
        database::instance().set_connection(app_connection, log_connection);
      }

      template<typename Key, typename Function>
      void register_function(Key func_name, Function func) { m_function_manager.register_function(func_name, func); }
      
      template<typename Key, typename Function>
      void register_inspector(Key func_name, Function func) { m_inspector.register_function(func_name, func); }

      void run()
      {
        // HERE
        // может в init перенести ???
        set_post();

        try 
        {
          hack::log(" ")("server listen:", var::HOST, var::PORT);
          listen(var::HOST, var::PORT);
        }
        catch(std::exception& e)
        {
          hack::error(" ")("error server run:", e.what());
        }
        catch(...)
        {
          hack::error()("SUPPER ERROR!!! GOOD LUCK MY FRIEND!!! :)");
        }
      }

    private:
      function_manager<transaction> m_function_manager;
      inspector<transaction> m_inspector;

    private:
      void set_CORS()
      {
        Options(R"(\*)", [](const auto& req, auto& res) { res.set_header("Allow", "POST, HEAD, OPTIONS"); });
        Options(var::API_URL, [](const auto& req, auto& res)
        {
          res.set_header("Access-Control-Allow-Origin", "*");
          res.set_header("Allow", "POST, HEAD, OPTIONS");
          res.set_header("Access-Control-Allow-Headers", "X-Requested-With, Content-Type, Accept, Origin, Authorization, X-server-function, X-token-auth");
          res.set_header("Access-Control-Allow-Methods", "OPTIONS, HEAD, POST");
        });
      }
      
      void set_post()
      {
        Post(var::API_URL, [&](const httplib::Request& req, httplib::Response& res) {
          res.set_header("Access-Control-Allow-Origin", "*");
          res.set_header("Access-Control-Allow-Headers", "*");
          res.set_header("Access-Control-Allow-Methods", "POST");
          res.set_header("Access-Control-Allow-Credentials", "false");

          transaction tr;

          try
          {
            tr.set_data(req);
            verification(m_inspector, m_function_manager, tr);
            execute(m_function_manager, tr);
          }
          catch(hack::exception& ex)
          {
            ex.transaction(tr);
            ex.commit<database>();
            ex.commit();
          }

          res.set_content(nlohmann::to_string(tr.m_data.m_result), var::HEADER_FLAG_JSON); 
        });
      }
  };
}


